from django.utils.translation import gettext as _

from jsonstore import IntegerField

from aleksis.apps.chronos import models as chronos_models
from aleksis.core import models as core_models

core_models.SchoolTerm.field(
    import_ref_untis=IntegerField(verbose_name=_("Untis import reference"), null=True, blank=True)
)
core_models.Person.field(
    import_ref_untis=IntegerField(verbose_name=_("Untis import reference"), null=True, blank=True)
)
core_models.Group.field(
    import_ref_untis=IntegerField(verbose_name=_("Untis import reference"), null=True, blank=True)
)

# Chronos models
chronos_models.ValidityRange.field(
    import_ref_untis=IntegerField(verbose_name=_("Untis import reference"), null=True, blank=True)
)
chronos_models.ValidityRange.field(
    school_year_untis=IntegerField(verbose_name=_("Untis school year ID"), null=True, blank=True)
)
chronos_models.ValidityRange.field(
    school_id_untis=IntegerField(verbose_name=_("Untis school id"), null=True, blank=True)
)
chronos_models.ValidityRange.field(
    version_id_untis=IntegerField(verbose_name=_("Untis version id"), null=True, blank=True)
)
chronos_models.Subject.field(
    import_ref_untis=IntegerField(verbose_name=_("Untis import reference"), null=True, blank=True)
)
chronos_models.Room.field(
    import_ref_untis=IntegerField(verbose_name=_("Untis import reference"), null=True, blank=True)
)
chronos_models.SupervisionArea.field(
    import_ref_untis=IntegerField(verbose_name=_("Untis import reference"), null=True, blank=True)
)
chronos_models.Lesson.field(
    lesson_id_untis=IntegerField(verbose_name=_("Lesson id in Untis"), null=True, blank=True)
)
chronos_models.Lesson.field(
    element_id_untis=IntegerField(
        verbose_name=_("Number of lesson element in Untis"), null=True, blank=True
    )
)
chronos_models.LessonPeriod.field(
    element_id_untis=IntegerField(
        verbose_name=_("Number of lesson element in Untis"), null=True, blank=True
    )
)
chronos_models.LessonSubstitution.field(
    import_ref_untis=IntegerField(verbose_name=_("Untis import reference"), null=True, blank=True)
)
chronos_models.SupervisionSubstitution.field(
    import_ref_untis=IntegerField(verbose_name=_("Untis import reference"), null=True, blank=True)
)
chronos_models.AbsenceReason.field(
    import_ref_untis=IntegerField(verbose_name=_("Untis import reference"), null=True, blank=True)
)
chronos_models.Absence.field(
    import_ref_untis=IntegerField(verbose_name=_("Untis import reference"), null=True, blank=True)
)
chronos_models.Event.field(
    import_ref_untis=IntegerField(verbose_name=_("UNTIS import reference"), null=True, blank=True)
)
chronos_models.Holiday.field(
    import_ref_untis=IntegerField(verbose_name=_("UNTIS import reference"), null=True, blank=True)
)
chronos_models.ExtraLesson.field(
    import_ref_untis=IntegerField(verbose_name=_("UNTIS import reference"), null=True, blank=True)
)



from typing import Optional

from django.db.models import Q, QuerySet
from django.utils.functional import classproperty

from aleksis.apps.untis.util.mysql.importers.terms import (
    get_future_terms_for_date,
    get_future_terms_for_date_query,
    get_terms,
    get_terms_for_date,
    get_terms_for_date_query,
)

from .util.mysql.main import untis_import_mysql as _untis_import_mysql


class ImportCommand:
    """A generic Untis import command."""

    name = None

    @classproperty
    def task_name(cls) -> str:  # noqa
        """Get the name for the related Celery task."""
        return f"untis_import_mysql_{cls.name}"

    @classmethod
    def get_terms(cls) -> Optional[QuerySet]:
        """Return which terms should be imported."""
        return None

    @classmethod
    def run(cls, background: bool = False, version: Optional[int] = None):
        """Run the import command (foreground/background)."""
        if background:
            from .tasks import TASKS

            task = TASKS[cls.task_name]
            task.delay(version=version)
        else:
            _untis_import_mysql(cls.get_terms(), version=version)


class CurrentImportCommand(ImportCommand):
    """Import data of current term from Untis."""

    name = "current"

    @classmethod
    def get_terms(cls) -> Optional[QuerySet]:
        return get_terms_for_date()


class FutureImportCommand(ImportCommand):
    """Import data of future terms from Untis."""

    name = "future"

    @classmethod
    def get_terms(cls) -> Optional[QuerySet]:
        return get_future_terms_for_date()


class AllImportCommand(ImportCommand):
    name = "all"


class CurrentNextImportCommand(ImportCommand):
    """Import data of the current and next term from Untis."""

    name = "current_next"

    @classmethod
    def get_terms(cls) -> Optional[QuerySet]:
        future_terms = get_future_terms_for_date()
        if future_terms.exists():
            future_term = future_terms.first()
            return get_terms().filter(
                get_terms_for_date_query()
                | Q(
                    school_id=future_term.school_id,
                    schoolyear_id=future_term.schoolyear_id,
                    version_id=future_term.version_id,
                    term_id=future_term.term_id,
                )
            )
        else:
            return get_terms_for_date()


class CurrentFutureImportCommand(ImportCommand):
    """Import data of the current and future terms from Untis."""

    name = "current_future"

    @classmethod
    def get_terms(cls) -> Optional[QuerySet]:
        terms = get_terms().filter(get_future_terms_for_date_query() | get_terms_for_date_query())
        return terms


COMMANDS_BY_NAME = {c.name: c for c in ImportCommand.__subclasses__()}
COMMANDS_BY_TASK_NAME = {c.task_name: c for c in ImportCommand.__subclasses__()}



import logging
from datetime import date, datetime
from typing import Any, Callable, Optional, Sequence, Union

from django.db.models import Model, QuerySet

from aleksis.apps.chronos.models import ValidityRange

DB_NAME = "untis"
UNTIS_DATE_FORMAT = "%Y%m%d"

TQDM_DEFAULTS = {
    "disable": None,
    "unit": "obj",
    "dynamic_ncols": True,
}

logger = logging.getLogger(__name__)


def run_using(obj: QuerySet) -> QuerySet:
    """Seed QuerySet with using() database from global DB_NAME."""
    return obj.using(DB_NAME)


def run_default_filter(
    validity_range: ValidityRange,
    qs: QuerySet,
    for_date: Optional[date] = None,
    filter_term: bool = True,
    filter_deleted: bool = True,
) -> QuerySet:
    """Add a default filter in order to select the correct term."""
    term_id, schoolyear_id, school_id, version_id = (
        validity_range.import_ref_untis,
        validity_range.school_year_untis,
        validity_range.school_id_untis,
        validity_range.version_id_untis,
    )

    qs = run_using(qs).filter(
        school_id=school_id,
        schoolyear_id=schoolyear_id,
        version_id=version_id,
    )

    if filter_term:
        qs = qs.filter(term_id=term_id)

    if filter_deleted:
        qs = qs.filter(deleted=0)

    return qs


def clean_array(seq: Sequence, conv: Callable[[Any], Any] = lambda el: el) -> Sequence:
    """Clean array.

    Convert a sequence using a converter function, stripping all
    elements that are boolean False after conversion.

    >>> clean_array(["a", "", "b"])
    ['a', 'b']

    >>> clean_array(["8", "", "12", "0"], int)
    [8, 12]
    """
    filtered = filter(lambda el: bool(el), map(lambda el: conv(el) if el else None, seq))
    return type(seq)(filtered)


def untis_split_first(s: str, conv: Callable[[Any], Any] = lambda el: el) -> Sequence:
    return clean_array(s.split(","), conv=conv)


def untis_split_second(s: str, conv: Callable[[Any], Any] = lambda el: el) -> Sequence:
    return clean_array(s.split("~"), conv=conv)


def untis_split_third(s: str, conv: Callable[[Any], Any] = lambda el: el) -> Sequence:
    return clean_array(s.split(";"), conv=conv)


def untis_date_to_date(untis: int) -> date:
    """Convert a Untis date to a python date."""
    return datetime.strptime(str(untis), UNTIS_DATE_FORMAT).date()


def date_to_untis_date(from_date: date) -> int:
    """Convert a python date to a Untis date."""
    return int(from_date.strftime(UNTIS_DATE_FORMAT))


def untis_colour_to_hex(colour: int) -> str:
    """Convert a numerical colour in BGR order to a standard hex RGB string."""
    # Convert Untis number to HEX
    hex_bgr = str(hex(colour))[2:].zfill(6)

    # Change BGR to RGB
    hex_rgb = hex_bgr[4:6] + hex_bgr[2:4] + hex_bgr[0:2]

    # Add html #
    return "#" + hex_rgb


def compare_m2m(a: Union[Sequence[Model], QuerySet], b: Union[Sequence[Model], QuerySet]) -> bool:
    """Compare if content of two m2m fields is equal."""
    return set(a) == set(b)


def connect_untis_fields(obj: Model, attr: str, limit: int) -> Sequence[str]:
    """Connect data from multiple DB fields.

    Untis splits structured data, like lists, as comma-separated string into
    multiple, numbered database fields, like:

      field1 = "This,is,a,nice"
      field2 = "list,of,words"

    This function joins these fields, then splits them into the original list.
    """
    all_data = []

    for i in range(1, limit + 1):
        attr_name = "{}{}".format(attr, i)
        raw_data = getattr(obj, attr_name, "")
        if raw_data:
            data = untis_split_first(raw_data)
            all_data += data

    return all_data


def get_first_weekday(time_periods_ref: dict) -> int:
    """Get first weekday from time periods reference."""
    return sorted(time_periods_ref.keys())[0]


def get_last_weekday(time_periods_ref: dict) -> int:
    """Get last weekday from time periods reference."""
    return sorted(time_periods_ref.keys())[-1]


def get_first_period(time_periods_ref: dict, weekday: int) -> int:
    """Get first period on a weekday from time periods reference."""
    return sorted(time_periods_ref[weekday].keys())[0]


def get_last_period(time_periods_ref: dict, weekday: int) -> int:
    """Get last period an a weekday from time periods reference."""
    return sorted(time_periods_ref[weekday].keys())[-1]


def move_weekday_to_range(time_periods_ref: dict, weekday: int) -> int:
    """Move weekday values into school week (e. g. saturday to friday)."""
    first_weekday = get_first_weekday(time_periods_ref)
    last_weekday = get_last_weekday(time_periods_ref)

    if weekday < first_weekday:
        weekday = first_weekday
    if weekday > last_weekday:
        weekday = last_weekday

    return weekday



from django.utils.translation import gettext_lazy as _

from dynamic_preferences.preferences import Section
from dynamic_preferences.types import BooleanPreference

from aleksis.core.registries import site_preferences_registry

untis_mysql = Section("untis_mysql", verbose_name=_("Untis: MySQL"))


@site_preferences_registry.register
class UpdateSubjects(BooleanPreference):
    section = untis_mysql
    name = "update_subjects"
    default = True
    verbose_name = _("Update values of existing subjects")


@site_preferences_registry.register
class UpdatePersonsShortName(BooleanPreference):
    section = untis_mysql
    name = "update_persons_short_name"
    default = False
    verbose_name = _("Update short name of existing persons")


@site_preferences_registry.register
class UpdatePersonsName(BooleanPreference):
    section = untis_mysql
    name = "update_persons_name"
    default = False
    verbose_name = _("Update name of existing persons")


@site_preferences_registry.register
class UpdateGroupsShortName(BooleanPreference):
    section = untis_mysql
    name = "update_groups_short_name"
    default = False
    verbose_name = _("Update short name of existing groups")


@site_preferences_registry.register
class UpdateGroupsName(BooleanPreference):
    section = untis_mysql
    name = "update_groups_name"
    default = False
    verbose_name = _("Update name of existing groups")


@site_preferences_registry.register
class DisambiguateGroupsName(BooleanPreference):
    section = untis_mysql
    name = "disambiguate_groups_name"
    default = True
    verbose_name = _("Disambiguate name of new groups")


@site_preferences_registry.register
class OverwriteGroupOwners(BooleanPreference):
    section = untis_mysql
    name = "overwrite_group_owners"
    verbose_name = _("Overwrite group owners")
    default = True


@site_preferences_registry.register
class UpdateRoomsName(BooleanPreference):
    section = untis_mysql
    name = "update_rooms_name"
    default = True
    verbose_name = _("Update name of existing rooms")


@site_preferences_registry.register
class UpdateSupervisionAreas(BooleanPreference):
    section = untis_mysql
    name = "update_supervision_areas"
    default = True
    verbose_name = _("Update existing supervision areas")


@site_preferences_registry.register
class UseCourseGroups(BooleanPreference):
    section = untis_mysql
    name = "use_course_groups"
    default = True
    verbose_name = _("Use course groups")
    help_text = _("Search course groups for every course instead of setting classes as groups.")


@site_preferences_registry.register
class CreateCourseGroups(BooleanPreference):
    section = untis_mysql
    name = "create_course_groups"
    default = True
    verbose_name = _("Create non-existing course groups")
    help_text = _("Only used if 'Use course groups' is enabled.")


@site_preferences_registry.register
class DataCheckNotFoundCourseGroups(BooleanPreference):
    section = untis_mysql
    name = "data_check_for_not_found_course_groups"
    default = True
    verbose_name = _("Register a data problem if a course group has been not found.")


@site_preferences_registry.register
class CourseGroupsFuzzyMatching(BooleanPreference):
    section = untis_mysql
    name = "course_groups_fuzzy_matching"
    default = False
    verbose_name = _("Match course groups by a subset of parent groups if no 100% match is found")
    help_text = _("Works only if 'Use course groups' is activated.")


@site_preferences_registry.register
class IgnoreIncompleteSubstitutions(BooleanPreference):
    section = untis_mysql
    name = "ignore_incomplete_substitutions"
    default = True
    verbose_name = _("Ignore incomplete substitutions")



from django.utils.translation import gettext_lazy as _

from aleksis.core.data_checks import DataCheck, IgnoreSolveOption


class CourseGroupNotFoundAndCreated(DataCheck):
    name = "untis_mysql_course_group_not_found_and_created"
    verbose_name = _(
        "Course groups created by the Untis import because no matching group has been found."
    )
    problem_name = _(
        "The Untis import created a new course group because no matching group has been found."
    )

    solve_options = {IgnoreSolveOption.name: IgnoreSolveOption}

    @classmethod
    def run_check_data(cls):
        pass


class CourseGroupNotFoundAndNotCreated(DataCheck):
    name = "untis_not_created_not_primary_source"
    verbose_name = _(
        "Course group not set by the Untis import because no matching group has been found."
    )
    problem_name = _(
        "The Untis import didn't set a course group "
        "for a lesson because no matching group has been found."
    )

    solve_options = {IgnoreSolveOption.name: IgnoreSolveOption}

    @classmethod
    def run_check_data(cls):
        pass



from django.core.management.base import BaseCommand

from ...commands import COMMANDS_BY_NAME


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "command", nargs="?", default="current", type=str, choices=list(COMMANDS_BY_NAME.keys())
        )
        parser.add_argument(
            "--background",
            action="store_true",
            help="Run import job in background using Celery",
        )
        parser.add_argument(
            "--plan-version",
            help="Select explicit Untis plan version",
        )

    def handle(self, *args, **options):
        command = COMMANDS_BY_NAME[options["command"]]
        background = options["background"]
        version = options.get("plan_version", None)
        command.run(background=background, version=version)



# Generated by Django 3.0.9 on 2020-08-20 13:42

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('untis', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='globalpermissions',
            options={'managed': False, 'permissions': (('assign_subjects_to_groups', 'Kann Fächer zu Gruppen zuzuordnen'),)},
        ),
    ]



import logging
from enum import Enum

from django.db.models import Q

from calendarweek import CalendarWeek
from tqdm import tqdm

from aleksis.apps.chronos import models as chronos_models
from aleksis.apps.chronos.models import ValidityRange
from aleksis.core.util.core_helpers import get_site_preferences

from .... import models as mysql_models
from ..util import (
    TQDM_DEFAULTS,
    date_to_untis_date,
    run_default_filter,
    untis_date_to_date,
    untis_split_first,
)

logger = logging.getLogger(__name__)


class SubstitutionFlag(Enum):
    CANCELLED = "E"
    CANCELLED_FOR_TEACHERS = "F"


def import_substitutions(
    validity_range: ValidityRange,
    teachers_ref,
    subjects_ref,
    rooms_ref,
    classes_ref,
    supervision_areas_ref,
    time_periods_ref,
):
    """Import substitutions."""
    subs = (
        run_default_filter(validity_range, mysql_models.Substitution.objects, filter_term=False)
        .filter(
            date__gte=date_to_untis_date(validity_range.date_start),
            date__lte=date_to_untis_date(validity_range.date_end),
        )
        .exclude(
            Q(flags__contains="N")
            | Q(flags__contains="b")
            | Q(flags__contains="F")
            | Q(flags__exact="g")
        )
        .order_by("substitution_id")
    )

    existing_subs = []
    for sub in tqdm(subs, desc="Import substitutions", **TQDM_DEFAULTS):
        # IDs
        sub_id = sub.substitution_id
        existing_subs.append(sub_id)

        lesson_id = sub.lesson_idsubst

        logger.info("Import substitution {}".format(sub_id))

        # Time
        date = untis_date_to_date(sub.date)
        weekday = date.weekday()
        week = CalendarWeek.from_date(date)
        period = sub.lesson

        # Supervision substitution?
        is_supervision_substitution = sub.corridor_id != 0

        # Cancellation?
        cancelled, cancelled_for_teachers = False, False
        if SubstitutionFlag.CANCELLED.value in sub.flags:
            cancelled = True
        elif SubstitutionFlag.CANCELLED_FOR_TEACHERS.value in sub.flags:
            cancelled_for_teachers = True

        # Comment
        comment = sub.text

        # Teacher
        if sub.teacher_idlessn != 0:
            teacher_old = teachers_ref[sub.teacher_idlessn]
        else:
            teacher_old = None

        teachers = []
        teacher_new = None
        if sub.teacher_idsubst != 0:
            teacher_new = teachers_ref[sub.teacher_idsubst]
            teachers = [teacher_new]

            if teacher_old is not None and teacher_new.id == teacher_old.id:
                teachers = []

        if not is_supervision_substitution:
            lesson_periods = chronos_models.LessonPeriod.objects.filter(
                lesson__validity=validity_range,
                lesson__lesson_id_untis=lesson_id,
                lesson__teachers=teacher_old,
                period__period=period,
                period__weekday=weekday,
            ).on_day(date)

            if lesson_periods.exists():
                lesson_period = lesson_periods[0]
                logger.info("  Matching lesson period found ({})".format(lesson_period))
            else:
                lesson_period = None

            # Subject
            subject_old = lesson_period.lesson.subject if lesson_period else None
            subject_new = None
            if sub.subject_idsubst != 0:
                subject_new = subjects_ref[sub.subject_idsubst]

                if subject_old and subject_old.id == subject_new.id:
                    subject_new = None

            if cancelled:
                subject_new = None

            # Room
            room_old = lesson_period.room if lesson_period else None
            room_new = None
            if sub.room_idsubst != 0:
                room_new = rooms_ref[sub.room_idsubst]

                if room_old is not None and room_old.id == room_new.id:
                    room_new = None

            # Classes
            classes = []
            class_ids = untis_split_first(sub.classids, conv=int)

            for id_ in class_ids:
                classes.append(classes_ref[id_])

            if get_site_preferences()["untis_mysql__ignore_incomplete_substitutions"]:
                # Skip incomplete substitutions
                if (
                    not room_new
                    and not subject_new
                    and not teachers
                    and not cancelled
                    and not cancelled_for_teachers
                    and not comment
                ):
                    continue

            if lesson_period:
                (substitution, created,) = chronos_models.LessonSubstitution.objects.get_or_create(
                    lesson_period=lesson_period, week=week.week, year=week.year
                )

                if created:
                    logger.info("  Substitution created")

                # Sync teachers
                substitution.teachers.set(teachers)
                logger.info("   Substitution teachers set")

                # Update values
                if (
                    substitution.subject != subject_new
                    or substitution.room != room_new
                    or substitution.cancelled != cancelled
                    or substitution.cancelled_for_teachers != cancelled_for_teachers
                    or substitution.comment != comment
                    or substitution.import_ref_untis != sub_id
                ):
                    substitution.subject = subject_new
                    substitution.room = room_new
                    substitution.cancelled = cancelled
                    substitution.cancelled_for_teachers = cancelled_for_teachers
                    substitution.comment = comment
                    substitution.import_ref_untis = sub_id
                    substitution.save()
                    logger.info("  Substitution updated")

            else:
                logger.info("  Extra lesson detected")
                time_period = time_periods_ref[date.weekday()][period]

                groups = [classes_ref[pk] for pk in untis_split_first(sub.classids, int)]

                room = room_old if not room_new and room_old else room_new
                subject = subject_old if not subject_new else subject_new
                teachers = [teacher_old] if not teacher_new else [teacher_new]
                if not teacher_new and not teacher_old:
                    teachers = []

                if not subject:
                    logger.warning("  Skip because missing subject")
                    continue

                (extra_lesson, created,) = (
                    chronos_models.ExtraLesson.objects.select_related(None)
                    .prefetch_related(None)
                    .update_or_create(
                        import_ref_untis=sub_id,
                        defaults={
                            "week": week.week,
                            "year": week.year,
                            "period": time_period,
                            "subject": subject,
                            "room": room,
                            "comment": comment,
                            "school_term": validity_range.school_term,
                        },
                    )
                )

                if created:
                    logger.info("  Extra lesson created")

                extra_lesson.teachers.set(teachers)
                extra_lesson.groups.set(groups)
        else:
            if teacher_new:
                logger.info("  Supervision substitution detected")

                # Supervision
                area_ref = supervision_areas_ref[sub.corridor_id]
                possible_supervisions = area_ref["supervisions"][weekday][period]

                supervision = None
                for possible_supervision in possible_supervisions:
                    if possible_supervision.teacher == teacher_old:
                        supervision = possible_supervision

                if supervision:
                    (
                        substitution,
                        created,
                    ) = chronos_models.SupervisionSubstitution.objects.get_or_create(
                        supervision=supervision,
                        date=date,
                        defaults={"teacher": teacher_new},
                    )

                    if created:
                        logger.info("  Supervision substitution created")

                    if (
                        substitution.teacher != teacher_new
                        or substitution.import_ref_untis != sub_id
                    ):
                        substitution.teacher = teacher_new
                        substitution.import_ref_untis = sub_id
                        substitution.save()
                        logger.info("  Supervision substitution updated")

    # Delete all no longer existing substitutions
    lesson_substitutions_in_range = chronos_models.LessonSubstitution.objects.within_dates(
        validity_range.date_start, validity_range.date_end
    )
    for s in lesson_substitutions_in_range:
        if s.import_ref_untis and s.import_ref_untis not in existing_subs:
            logger.info("Substitution {} deleted".format(s.id))
            s.delete()

    lesson_substitutions_in_range.exclude(lesson_period__lesson__validity=validity_range).delete()

    # Delete all no longer existing extra lessons
    for s in chronos_models.ExtraLesson.objects.within_dates(
        validity_range.date_start, validity_range.date_end
    ):
        if s.import_ref_untis and s.import_ref_untis not in existing_subs:
            logger.info("Extra lesson {} deleted".format(s.id))
            s.delete()

    # Delete all no longer existing supervision substitutions
    supervision_substitutions_in_range = chronos_models.SupervisionSubstitution.objects.filter(
        date__gte=validity_range.date_start, date__lte=validity_range.date_end
    )
    for s in supervision_substitutions_in_range:
        if s.import_ref_untis and s.import_ref_untis not in existing_subs:
            logger.info("Supervision substitution {} deleted".format(s.id))
            s.delete()

    supervision_substitutions_in_range.exclude(supervision__validity=validity_range).delete()



import logging
import re
from datetime import time
from enum import Enum
from typing import Dict

from tqdm import tqdm

from aleksis.apps.chronos import models as chronos_models
from aleksis.apps.chronos.models import ValidityRange
from aleksis.core import models as core_models
from aleksis.core.util.core_helpers import get_site_preferences

from .... import models as mysql_models
from ..util import (
    TQDM_DEFAULTS,
    connect_untis_fields,
    run_default_filter,
    untis_colour_to_hex,
    untis_split_first,
)

logger = logging.getLogger(__name__)


class CommonDataId(Enum):
    PERIOD = 40


def import_subjects(validity_range: ValidityRange) -> Dict[int, chronos_models.Subject]:
    """Import subjects."""
    subjects_ref = {}

    # Get subjects
    subjects = run_default_filter(validity_range, mysql_models.Subjects.objects, filter_term=False)

    for subject in tqdm(subjects, desc="Import subjects", **TQDM_DEFAULTS):
        # Check if needed data are provided
        if not subject.name:
            raise RuntimeError(
                "Subject ID {}: Cannot import subject without short name.".format(
                    subject.subject_id
                )
            )

        # Build values
        short_name = subject.name[:10]
        name = subject.longname if subject.longname else short_name
        colour_fg = untis_colour_to_hex(subject.forecolor)
        colour_bg = untis_colour_to_hex(subject.backcolor)
        import_ref = subject.subject_id

        logger.info("Import subject {} …".format(short_name))

        # Get or create subject object by short name
        new_subject, created = chronos_models.Subject.objects.get_or_create(
            short_name=short_name,
            defaults={
                "name": name,
                "colour_fg": colour_fg,
                "colour_bg": colour_bg,
                "import_ref_untis": import_ref,
            },
        )

        if created:
            logger.info("  New subject created")

        # Force sync
        changed = False
        if get_site_preferences()["untis_mysql__update_subjects"] and (
            new_subject.name != name
            or new_subject.colour_fg != colour_fg
            or new_subject.colour_bg != colour_bg
        ):
            new_subject.name = name
            new_subject.colour_fg = untis_colour_to_hex(subject.forecolor)
            new_subject.colour_bg = untis_colour_to_hex(subject.backcolor)
            changed = True

            logger.info("  Name, foreground and background colour updated")

        if new_subject.import_ref_untis != import_ref:
            new_subject.import_ref_untis = import_ref
            changed = True

            logger.info("  Import reference updated")

        if changed:
            new_subject.save()

        subjects_ref[import_ref] = new_subject

    return subjects_ref


def import_teachers(validity_range: ValidityRange) -> Dict[int, core_models.Person]:
    """Import teachers."""
    teachers_ref = {}

    # Get teachers
    teachers = run_default_filter(validity_range, mysql_models.Teacher.objects)

    for teacher in tqdm(teachers, desc="Import teachers", **TQDM_DEFAULTS):
        # Check if needed data are provided
        if not teacher.name:
            raise RuntimeError(
                "Teacher ID {}: Cannot import teacher without short name.".format(
                    teacher.teacher_id
                )
            )

        # Build values
        short_name = teacher.name
        first_name = teacher.firstname if teacher.firstname else "?"
        last_name = teacher.longname if teacher.longname else teacher.name
        import_ref = teacher.teacher_id

        logger.info("Import teacher {} (as person) …".format(short_name))

        try:
            new_teacher = core_models.Person.objects.get(short_name__iexact=short_name)
        except core_models.Person.DoesNotExist:
            new_teacher = core_models.Person.objects.create(
                short_name=short_name,
                first_name=first_name,
                last_name=last_name,
                import_ref_untis=import_ref,
            )
            logger.info("  New person created")

        changed = False
        if get_site_preferences()["untis_mysql__update_persons_name"] and (
            new_teacher.first_name != first_name or new_teacher.last_name != last_name
        ):
            new_teacher.first_name = first_name
            new_teacher.last_name = last_name
            changed = True
            logger.info("  First and last name updated")

        if (
            get_site_preferences()["untis_mysql__update_persons_short_name"]
            and new_teacher.short_name != short_name
        ):
            new_teacher.short_name = short_name
            changed = True
            logger.info("  Short name updated")

        if new_teacher.import_ref_untis != import_ref:
            new_teacher.import_ref_untis = import_ref
            changed = True
            logger.info("  Import reference updated")

        if changed:
            new_teacher.save()

        teachers_ref[teacher.teacher_id] = new_teacher

    return teachers_ref


def import_classes(
    validity_range: ValidityRange, teachers_ref: Dict[int, core_models.Person]
) -> Dict[int, core_models.Group]:
    """Import classes."""
    classes_ref = {}

    # Get classes
    course_classes = run_default_filter(
        validity_range, mysql_models.Class.objects, filter_term=True
    )

    for class_ in tqdm(course_classes, desc="Import classes", **TQDM_DEFAULTS):
        # Check if needed data are provided
        if not class_.name:
            raise RuntimeError(
                "Class ID {}: Cannot import class without short name.".format(class_.teacher_id)
            )

        # Build values
        short_name = class_.name[:16]
        name = class_.longname if class_.longname else short_name
        teacher_ids = untis_split_first(class_.teacherids, int) if class_.teacherids else []
        owners = [teachers_ref[t_id] for t_id in teacher_ids]
        import_ref = class_.class_id

        logger.info("Import class {} (as group) …".format(short_name))

        try:
            new_group = core_models.Group.objects.get(
                short_name__iexact=short_name,
                school_term__in=[None, validity_range.school_term],
            )
        except core_models.Group.DoesNotExist:
            # Determine collisions of long name
            while core_models.Group.objects.filter(
                name=name, school_term=validity_range.school_term
            ).exists():
                if not get_site_preferences()["untis_mysql__disambiguate_groups_name"]:
                    # Do not try to disambiguate it not enabled
                    break

                suffix = f" ({short_name})"
                if name.endswith(suffix):
                    # First try, add a counter
                    name = f"{name} (2)"
                else:
                    # Second or more tries, determine last counter, if any
                    match = re.match(r"^(.*)(\(\d+\))$", name)
                    if match:
                        # Counter found, increase
                        prefix = match.group(1)
                        counter = int(match.group(2)) + 1
                        name = f"{prefix}({counter})"
                    else:
                        # Counter not found, add one
                        name = f"{name} (2)"

            new_group = core_models.Group.objects.create(
                short_name=short_name,
                name=name,
                import_ref_untis=import_ref,
                school_term=validity_range.school_term,
            )
            logger.info("  New group created")

        changed = False

        if (
            get_site_preferences()["untis_mysql__update_groups_short_name"]
            and new_group.short_name != short_name
        ):
            new_group.short_name = short_name
            changed = True
            logger.info("  Short name updated")

        if get_site_preferences()["untis_mysql__update_groups_name"] and new_group.name != name:
            new_group.name = name
            changed = True
            logger.info("  Name updated")

        if new_group.import_ref_untis != import_ref:
            new_group.import_ref_untis = import_ref
            changed = True
            logger.info("  Import reference updated")

        if new_group.school_term != validity_range.school_term:
            new_group.school_term = validity_range.school_term
            changed = True
            logger.info("  School term updated")

        if changed:
            new_group.save()

        if get_site_preferences()["untis_mysql__overwrite_group_owners"]:
            new_group.owners.set(owners)
            logger.info("  Group owners set")
        else:
            new_group.owners.add(*owners)
            logger.info("  Group owners added")

        classes_ref[class_.class_id] = new_group

    return classes_ref


def import_rooms(validity_range: ValidityRange) -> Dict[int, chronos_models.Room]:
    """Import rooms."""
    ref = {}

    # Get rooms
    rooms = run_default_filter(validity_range, mysql_models.Room.objects)

    for room in tqdm(rooms, desc="Import rooms", **TQDM_DEFAULTS):
        if not room.name:
            raise RuntimeError(
                "Room ID {}: Cannot import room without short name.".format(room.room_id)
            )

        # Build values
        short_name = room.name[:10]
        name = room.longname[:30] if room.longname else short_name
        import_ref = room.room_id

        logger.info("Import room {} …".format(short_name))

        new_room, created = chronos_models.Room.objects.get_or_create(
            short_name=short_name,
            defaults={"name": name, "import_ref_untis": import_ref},
        )

        if created:
            logger.info("  New room created")

        changed = False

        if get_site_preferences()["untis_mysql__update_rooms_name"] and new_room.name != name:
            new_room.name = name
            changed = True
            logger.info("  Name updated")

        if new_room.import_ref_untis != import_ref:
            new_room.import_ref_untis = import_ref
            changed = True
            logger.info("  Import reference updated")

        if changed:
            new_room.save()

        ref[import_ref] = new_room

    return ref


def import_supervision_areas(
    validity_range: ValidityRange, breaks_ref, teachers_ref
) -> Dict[int, chronos_models.SupervisionArea]:
    """Import supervision areas."""
    ref = {}

    # Get supervision areas
    areas = run_default_filter(validity_range, mysql_models.Corridor.objects, filter_term=False)

    for area in tqdm(areas, desc="Import supervision areas", **TQDM_DEFAULTS):
        if not area.name:
            raise RuntimeError(
                "Supervision area ID {}: Cannot import supervision area without short name.".format(
                    area.corridor_id
                )
            )

        short_name = area.name[:10]
        name = area.longname[:50] if area.longname else short_name
        colour_fg = untis_colour_to_hex(area.forecolor)
        colour_bg = untis_colour_to_hex(area.backcolor)
        import_ref = area.corridor_id

        logger.info("Import supervision area {} …".format(short_name))

        new_area, created = chronos_models.SupervisionArea.objects.get_or_create(
            short_name=short_name,
            defaults={
                "name": name,
                "colour_fg": colour_fg,
                "colour_bg": colour_bg,
                "import_ref_untis": import_ref,
            },
        )

        if created:
            logger.info("  New supervision area created")

        changed = False

        if get_site_preferences()["untis_mysql__update_supervision_areas"] and (
            new_area.name != new_area.name
            or new_area.colour_fg != colour_fg
            or new_area.colour_bg != colour_bg
        ):
            new_area.name = name
            new_area.colour_fg = colour_fg
            new_area.colour_bg = colour_bg
            changed = True

            logger.info("  Name, foreground and background colour updated")

        if new_area.import_ref_untis != import_ref:
            new_area.import_ref_untis = import_ref
            changed = True
            logger.info("  Import reference updated")

        if changed:
            new_area.save()

        logger.info("  Import supervisions for this area")

        # Parse raw data
        raw_supervisions = connect_untis_fields(area, "breaksupervision", 16)

        supervisions_ref = {}
        for raw_supervision in raw_supervisions:
            # Split more and get teacher id
            raw_supervision_2 = raw_supervision.split("~")
            teacher_id = int(raw_supervision_2[1])
            term_id = int(raw_supervision_2[0])

            if term_id != validity_range.import_ref_untis:
                continue

            if teacher_id in teachers_ref:
                # Get weekday, period after break and teacher
                weekday = int(raw_supervision_2[2]) - 1
                period_after_break = int(raw_supervision_2[3])
                teacher = teachers_ref[teacher_id]

                logger.info(
                    "Import supervision on weekday {} before the {}. period (teacher {})".format(
                        weekday, period_after_break, teacher
                    )
                )

                # Get or create
                new_supervision, created = new_area.supervisions.get_or_create(
                    validity=validity_range,
                    break_item=breaks_ref[weekday][period_after_break],
                    teacher=teacher,
                )

                # Log
                if created:
                    logger.info("      New supervision created")

                # Save supervisions in reference dict
                if weekday not in supervisions_ref:
                    supervisions_ref[weekday] = {}
                if period_after_break not in supervisions_ref[weekday]:
                    supervisions_ref[weekday][period_after_break] = []
                supervisions_ref[weekday][period_after_break].append(new_supervision)

        for supervision in new_area.supervisions.filter(validity=validity_range):
            delete = True

            # Get weekday and period after break
            weekday = supervision.break_item.weekday
            period_after_break = supervision.break_item.before_period_number

            # Delete supervision if no longer existing
            if weekday in supervisions_ref:
                if period_after_break in supervisions_ref[weekday]:
                    if supervision in supervisions_ref[weekday][period_after_break]:
                        delete = False

            if delete:
                supervision.delete()
                logger.info("    Supervision {} deleted".format(supervision))

        ref[import_ref] = {"area": new_area, "supervisions": supervisions_ref}

    return ref


def import_time_periods(
    validity_range: ValidityRange,
) -> Dict[int, Dict[int, chronos_models.TimePeriod]]:
    """Import time periods an breaks."""
    times = (
        run_default_filter(validity_range, mysql_models.Commondata.objects, filter_term=False)
        .filter(id=30)
        .order_by("number")
    )

    times_ref = {}
    for time_ in tqdm(times, desc="Import times", **TQDM_DEFAULTS):
        period = time_.number

        # Extract time
        start_time = time(time_.fieldbyte1, time_.fieldbyte2)
        end_time = time(time_.fieldbyte3, time_.fieldbyte4)

        times_ref[period] = (start_time, end_time)

    periods = (
        run_default_filter(validity_range, mysql_models.Commondata.objects, filter_term=False)
        .filter(id=CommonDataId.PERIOD.value)
        .order_by("number", "number1")
    )

    time_periods_ref = {}
    for time_period in tqdm(periods, desc="Import time periods", **TQDM_DEFAULTS):
        weekday = time_period.number - 1
        period = time_period.number1
        start_time = times_ref[period][0]
        end_time = times_ref[period][1]

        logger.info("Import time period on weekday {} in the {}. period".format(weekday, period))

        new_time_period, created = chronos_models.TimePeriod.objects.get_or_create(
            weekday=weekday,
            period=period,
            validity=validity_range,
            defaults={"time_start": start_time, "time_end": end_time},
        )

        if created:
            logger.info("  New time period created")

        if new_time_period.time_start != start_time or new_time_period.time_end != end_time:
            new_time_period.time_start = start_time
            new_time_period.time_end = end_time
            new_time_period.save()
            logger.info("  Time period updated")

        # Build index with time periods
        if weekday not in time_periods_ref:
            time_periods_ref[weekday] = {}
        time_periods_ref[weekday][period] = new_time_period

    return time_periods_ref


def import_breaks(
    validity_range: ValidityRange,
    time_periods_ref: Dict[int, Dict[int, chronos_models.TimePeriod]],
) -> Dict[int, Dict[int, chronos_models.Break]]:
    # Build breaks for all weekdays
    breaks_ref = {}
    for weekday, time_periods in tqdm(
        time_periods_ref.items(), desc="Import breaks (weekday)", **TQDM_DEFAULTS
    ):
        breaks_ref[weekday] = {}

        # Add None two times in order to create breaks before first lesson and after last lesson
        time_periods_for_breaks = [None] + list(time_periods.values()) + [None]
        for i, time_period in tqdm(
            enumerate(time_periods_for_breaks), desc="Import breaks (period)", **TQDM_DEFAULTS
        ):
            # If last item (None) is reached, no further break must be created
            if i + 1 == len(time_periods_for_breaks):
                break

            after_period = time_period
            before_period = time_periods_for_breaks[i + 1]

            short_name = "{}: {}./{}.".format(
                weekday,
                after_period.period if after_period else "-",
                before_period.period if before_period else "-",
            )

            logger.info("Generate break {}".format(short_name))

            new_break, created = chronos_models.Break.objects.get_or_create(
                after_period=after_period,
                before_period=before_period,
                validity=validity_range,
                defaults={"short_name": short_name, "name": short_name},
            )

            if created:
                logger.info("  New break created")

            # Save index with lesson after break
            next_period = new_break.before_period_number
            breaks_ref[weekday][next_period] = new_break

    return breaks_ref


def import_absence_reasons(
    validity_range: ValidityRange,
) -> Dict[int, chronos_models.AbsenceReason]:
    """Import absence reasons."""
    ref = {}

    # Get reasons
    reasons = run_default_filter(
        validity_range, mysql_models.Absencereason.objects, filter_term=False
    )

    for reason in tqdm(reasons, desc="Import absence reasons", **TQDM_DEFAULTS):
        if not reason.name:
            raise RuntimeError(
                "Absence reason ID {}: Cannot import absence reason without short name.".format(
                    reason.absence_reason_id
                )
            )

        # Build values
        short_name = reason.name
        name = reason.longname if reason.longname else short_name
        import_ref = reason.absence_reason_id

        logger.info("Import absence reason {} …".format(short_name))

        new_reason, created = chronos_models.AbsenceReason.objects.get_or_create(
            import_ref_untis=import_ref,
            defaults={"short_name": short_name, "name": name},
        )

        if created:
            logger.info("  New absence reason created")

        changed = False

        if new_reason.short_name != short_name or new_reason.name != name:
            new_reason.short_name = short_name
            new_reason.name = name
            changed = True
            logger.info("  Short name and name updated")

        if changed:
            new_reason.save()

        ref[import_ref] = new_reason

    return ref
